@@ -45,7 +45,7 @@ __all__ = [  # noqa
     "start_transaction",
     "trace",
     "monitor",
-    "_experimental_logger",
+    "logger",
 ]
 
 # Initialize the debug support after everything is loaded
